// Copyright 2021 PingCAP, Inc.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// See the License for the specific language governing permissions and
// limitations under the License.

package processor

import (
	"context"
	"testing"
	"time"

	"github.com/pingcap/failpoint"
	"github.com/stretchr/testify/mock"
	"github.com/stretchr/testify/require"
	"github.com/tikv/migration/cdc/cdc/model"
	pscheduler "github.com/tikv/migration/cdc/cdc/scheduler"
	cdcContext "github.com/tikv/migration/cdc/pkg/context"
	"github.com/tikv/migration/cdc/pkg/etcd"
	"github.com/tikv/migration/cdc/pkg/p2p"
	"github.com/tikv/migration/cdc/pkg/version"
	"go.etcd.io/etcd/clientv3"
	"go.etcd.io/etcd/mvcc/mvccpb"
)

const (
	// The purpose of these tests is to test the correctness
	// of the agent alone. So one node for the agent and the
	// other for the scheduler is enough.
	agentTestMockNodeNum = 2

	ownerCaptureID     = "capture-0"
	processorCaptureID = "capture-1"
)

// agentTestSuite is a test suite for the agent (the component of the processor).
// NOTE these tests are actually an integration test for the agent and the p2p package.
// TODO add a real unit test with mock components for the agent alone,
// which might require refactoring some existing components.
type agentTestSuite struct {
	cluster      *p2p.MockCluster
	etcdClient   *clientv3.Client
	etcdKVClient *mockEtcdKVClient

	keyspanExecutor    *pscheduler.MockKeySpanExecutor
	dispatchResponseCh chan *model.DispatchKeySpanResponseMessage
	syncCh             chan *model.SyncMessage
	checkpointCh       chan *model.CheckpointMessage

	ownerMessageClient *p2p.MessageClient

	ctx    context.Context
	cdcCtx cdcContext.Context
	cancel context.CancelFunc
}

func newAgentTestSuite(t *testing.T) *agentTestSuite {
	ctx, cancel := context.WithCancel(context.Background())
	etcdCli, KVCli := newMockEtcdClientForAgentTests(ctx)

	cluster := p2p.NewMockCluster(t, agentTestMockNodeNum)
	ownerMessageServer := cluster.Nodes[ownerCaptureID].Server

	ownerMessageClient := cluster.Nodes[ownerCaptureID].Router.GetClient(processorCaptureID)
	require.NotNil(t, ownerMessageClient)

	dispatchResponseCh := make(chan *model.DispatchKeySpanResponseMessage, 1)
	_, err := ownerMessageServer.SyncAddHandler(ctx, model.DispatchKeySpanResponseTopic("cf-1"),
		&model.DispatchKeySpanResponseMessage{},
		func(senderID string, msg interface{}) error {
			require.Equal(t, processorCaptureID, senderID)
			require.IsType(t, &model.DispatchKeySpanResponseMessage{}, msg)
			dispatchResponseCh <- msg.(*model.DispatchKeySpanResponseMessage)
			return nil
		},
	)
	require.NoError(t, err)

	syncCh := make(chan *model.SyncMessage, 1)
	_, err = ownerMessageServer.SyncAddHandler(ctx, model.SyncTopic("cf-1"),
		&model.SyncMessage{},
		func(senderID string, msg interface{}) error {
			require.Equal(t, processorCaptureID, senderID)
			require.IsType(t, &model.SyncMessage{}, msg)
			syncCh <- msg.(*model.SyncMessage)
			return nil
		},
	)
	require.NoError(t, err)

	checkpointCh := make(chan *model.CheckpointMessage, 1)
	_, err = ownerMessageServer.SyncAddHandler(ctx, model.CheckpointTopic("cf-1"),
		&model.CheckpointMessage{},
		func(senderID string, msg interface{}) error {
			require.Equal(t, processorCaptureID, senderID)
			require.IsType(t, &model.CheckpointMessage{}, msg)
			checkpointCh <- msg.(*model.CheckpointMessage)
			return nil
		},
	)
	require.NoError(t, err)

	return &agentTestSuite{
		cluster:      cluster,
		etcdClient:   etcdCli,
		etcdKVClient: KVCli,

		dispatchResponseCh: dispatchResponseCh,
		syncCh:             syncCh,
		checkpointCh:       checkpointCh,

		ownerMessageClient: ownerMessageClient,

		ctx:    ctx,
		cancel: cancel,
	}
}

func (s *agentTestSuite) CreateAgent(t *testing.T) (*agentImpl, error) {
	cdcEtcdClient := etcd.NewCDCEtcdClient(s.ctx, s.etcdClient)
	messageServer := s.cluster.Nodes["capture-1"].Server
	messageRouter := s.cluster.Nodes["capture-1"].Router
	s.keyspanExecutor = pscheduler.NewMockKeySpanExecutor(t)

	ctx := cdcContext.NewContext(s.ctx, &cdcContext.GlobalVars{
		EtcdClient:    &cdcEtcdClient,
		MessageServer: s.cluster.Nodes[processorCaptureID].Server,
		MessageRouter: s.cluster.Nodes[processorCaptureID].Router,
	})
	s.cdcCtx = ctx

	ret, err := newAgent(ctx, messageServer, messageRouter, s.keyspanExecutor, "cf-1")
	if err != nil {
		return nil, err
	}

	return ret.(*agentImpl), nil
}

func (s *agentTestSuite) Close() {
	s.cancel()
	s.cluster.Close()
}

// newMockEtcdClientForAgentTests returns a mock Etcd client.
// NOTE: The mock client does not have any useful internal logic.
// It only supports GET operations and any output should be supplied by
// calling the mock.Mock methods embedded in the mock client.
func newMockEtcdClientForAgentTests(ctx context.Context) (*clientv3.Client, *mockEtcdKVClient) {
	cli := clientv3.NewCtxClient(ctx)
	mockKVCli := &mockEtcdKVClient{}
	cli.KV = mockKVCli
	return cli, mockKVCli
}

type mockEtcdKVClient struct {
	clientv3.KV // embeds a null implementation of the Etcd KV client
	mock.Mock
}

func (c *mockEtcdKVClient) Get(ctx context.Context, key string, opts ...clientv3.OpOption) (*clientv3.GetResponse, error) {
	args := c.Called(ctx, key, opts)
	resp := (*clientv3.GetResponse)(nil)
	if args.Get(0) != nil {
		resp = args.Get(0).(*clientv3.GetResponse)
	}
	return resp, args.Error(1)
}

func TestAgentBasics(t *testing.T) {
	suite := newAgentTestSuite(t)
	defer suite.Close()

	suite.etcdKVClient.On("Get", mock.Anything, etcd.CaptureOwnerKey, mock.Anything).Return(&clientv3.GetResponse{
		Kvs: []*mvccpb.KeyValue{
			{
				Key:         []byte(etcd.CaptureOwnerKey),
				Value:       []byte(ownerCaptureID),
				ModRevision: 1,
			},
		},
	}, nil).Once()

	// Test Point 1: Create an agent.
	agent, err := suite.CreateAgent(t)
	require.NoError(t, err)

	// Test Point 2: First tick should sync the SyncMessage.
	err = agent.Tick(suite.cdcCtx)
	require.NoError(t, err)

	select {
	case <-suite.ctx.Done():
		require.Fail(t, "context should not be canceled")
	case syncMsg := <-suite.syncCh:
		require.Equal(t, &model.SyncMessage{
			ProcessorVersion: version.ReleaseSemver(),
			Running:          nil,
			Adding:           nil,
			Removing:         nil,
		}, syncMsg)
	}

	_, err = suite.ownerMessageClient.SendMessage(suite.ctx, model.DispatchKeySpanTopic("cf-1"), &model.DispatchKeySpanMessage{
		OwnerRev: 1,
		ID:       1,
		IsDelete: false,
	})
	require.NoError(t, err)

	// Test Point 3: Accept an incoming DispatchKeySpanMessage, and the AddKeySpan method in KeySpanExecutor can return false.
	suite.keyspanExecutor.On("AddKeySpan", mock.Anything, model.KeySpanID(1)).Return(false, nil).Once()
	suite.keyspanExecutor.On("AddKeySpan", mock.Anything, model.KeySpanID(1)).Return(true, nil).Run(
		func(_ mock.Arguments) {
			delete(suite.keyspanExecutor.Adding, 1)
			suite.keyspanExecutor.Running[1] = struct{}{}
		}).Once()
	suite.keyspanExecutor.On("GetCheckpoint").Return(model.Ts(1000), model.Ts(1000))

	require.Eventually(t, func() bool {
		err = agent.Tick(suite.cdcCtx)
		require.NoError(t, err)
		if len(suite.keyspanExecutor.Running) != 1 {
			return false
		}
		select {
		case <-suite.ctx.Done():
			require.Fail(t, "context should not be canceled")
		case msg := <-suite.checkpointCh:
			require.Equal(t, &model.CheckpointMessage{
				CheckpointTs: model.Ts(1000),
				ResolvedTs:   model.Ts(1000),
			}, msg)
			return true
		default:
		}
		return false
	}, 5*time.Second, 100*time.Millisecond)

	suite.keyspanExecutor.AssertExpectations(t)
	suite.keyspanExecutor.ExpectedCalls = nil
	suite.keyspanExecutor.Calls = nil

	// Test Point 4: Accept an incoming DispatchKeySpanMessage, and the AddKeySpan method in KeySpanExecutor can return true.
	err = agent.Tick(suite.cdcCtx)
	require.NoError(t, err)

	suite.keyspanExecutor.AssertExpectations(t)
	suite.keyspanExecutor.ExpectedCalls = nil
	suite.keyspanExecutor.Calls = nil

	require.Eventually(t, func() bool {
		select {
		case <-suite.ctx.Done():
			return false
		case msg := <-suite.dispatchResponseCh:
			require.Equal(t, &model.DispatchKeySpanResponseMessage{
				ID: 1,
			}, msg)
			return true
		default:
		}

		err = agent.Tick(suite.cdcCtx)
		require.NoError(t, err)
		return false
	}, time.Second*3, time.Millisecond*10)

	// Test Point 5: Close the agent.
	err = agent.Close()
	require.NoError(t, err)

	// double close
	err = agent.Close()
	require.NoError(t, err)
}

func TestAgentNoOwnerAtStartUp(t *testing.T) {
	suite := newAgentTestSuite(t)
	defer suite.Close()

	// Empty response implies no owner.
	suite.etcdKVClient.On("Get", mock.Anything, etcd.CaptureOwnerKey, mock.Anything).
		Return(&clientv3.GetResponse{}, nil).Once()

	// Test Point 1: Create an agent.
	agent, err := suite.CreateAgent(t)
	require.NoError(t, err)

	// Test Point 2: First ticks should not panic
	for i := 0; i < 10; i++ {
		err = agent.Tick(suite.cdcCtx)
		require.NoError(t, err)
	}

	// Test Point 3: Agent should process the Announce message.
	_, err = suite.ownerMessageClient.SendMessage(suite.ctx, model.AnnounceTopic("cf-1"), &model.AnnounceMessage{
		OwnerRev:     1,
		OwnerVersion: version.ReleaseSemver(),
	})
	require.NoError(t, err)

	require.Eventually(t, func() bool {
		err := agent.Tick(suite.cdcCtx)
		require.NoError(t, err)
		select {
		case <-suite.ctx.Done():
			require.Fail(t, "context should not be canceled")
		case syncMsg := <-suite.syncCh:
			require.Equal(t, &model.SyncMessage{
				ProcessorVersion: version.ReleaseSemver(),
				Running:          nil,
				Adding:           nil,
				Removing:         nil,
			}, syncMsg)
			return true
		default:
		}
		return false
	}, 5*time.Second, 100*time.Millisecond)

	// Test Point 4: Close the agent.
	err = agent.Close()
	require.NoError(t, err)

	// double close
	err = agent.Close()
	require.NoError(t, err)
}

func TestAgentTolerateClientClosed(t *testing.T) {
	suite := newAgentTestSuite(t)
	defer suite.Close()

	suite.etcdKVClient.On("Get", mock.Anything, etcd.CaptureOwnerKey, mock.Anything).Return(&clientv3.GetResponse{
		Kvs: []*mvccpb.KeyValue{
			{
				Key:         []byte(etcd.CaptureOwnerKey),
				Value:       []byte(ownerCaptureID),
				ModRevision: 1,
			},
		},
	}, nil).Once()

	// Test Point 1: Create an agent.
	agent, err := suite.CreateAgent(t)
	require.NoError(t, err)

	_ = failpoint.Enable("github.com/tikv/migration/cdc/pkg/p2p/ClientInjectClosed", "5*return(true)")
	defer func() {
		_ = failpoint.Disable("github.com/tikv/migration/cdc/pkg/p2p/ClientInjectClosed")
	}()

	// Test Point 2: We should tolerate the error ErrPeerMessageClientClosed
	for i := 0; i < 6; i++ {
		err = agent.Tick(suite.cdcCtx)
		require.NoError(t, err)
	}

	select {
	case <-suite.ctx.Done():
		require.Fail(t, "context should not be canceled")
	case syncMsg := <-suite.syncCh:
		require.Equal(t, &model.SyncMessage{
			ProcessorVersion: version.ReleaseSemver(),
			Running:          nil,
			Adding:           nil,
			Removing:         nil,
		}, syncMsg)
	}
}
